/*
 * Copyright (C) 2007-2009 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file blob.c Macros and functions for blob handling. */

#include <fnmatch.h>
#include "para.h"
#include "error.h"
#include "string.h"
#include "afh.h"
#include "afs.h"
#include "net.h"
#include "ipc.h"

static struct osl_column_description blob_cols[] = {
	[BLOBCOL_ID] = {
		.storage_type = OSL_MAPPED_STORAGE,
		.storage_flags = OSL_RBTREE | OSL_UNIQUE | OSL_FIXED_SIZE,
		.name = "id",
		.data_size = 4,
		.compare_function = uint32_compare
	},
	[BLOBCOL_NAME] = {
		.storage_type = OSL_MAPPED_STORAGE,
		.storage_flags = OSL_RBTREE | OSL_UNIQUE,
		.name = "name",
		.compare_function = string_compare
	},
	[BLOBCOL_DEF] = {
		.storage_type = OSL_DISK_STORAGE,
		.storage_flags = 0,
		.name = "definition"
	}
};

/** \cond doxygen isn't smart enough to recognize these */
INIT_BLOB_TABLE(lyrics);
INIT_BLOB_TABLE(images);
INIT_BLOB_TABLE(moods);
INIT_BLOB_TABLE(playlists);
/** \endcond */

/** Flags that may be passed to the \p ls functions of each blob  type. */
enum blob_ls_flags {
	/** List both id and name. */
	BLOB_LS_FLAG_LONG = 1,
	/** Reverse sort order. */
	BLOB_LS_FLAG_REVERSE = 2,
	/** Sort by id instead of name. */
	BLOB_LS_FLAG_SORT_BY_ID = 4,
};

/** Structure passed to the \p print_blob function. */
struct lsblob_action_data {
	/** The flags given at the command line. */
	uint32_t flags;
	/** Message buffer. */
	struct para_buffer pb;
};

static int print_blob(struct osl_table *table, struct osl_row *row,
		const char *name, void *data)
{
	struct lsblob_action_data *lbad = data;
	struct osl_object obj;
	uint32_t id;
	int ret;

	if (!(lbad->flags & BLOB_LS_FLAG_LONG))
		return para_printf(&lbad->pb, "%s\n", name);
	ret = osl_get_object(table, row, BLOBCOL_ID, &obj);
	if (ret < 0) {
		para_printf(&lbad->pb, "%s: %s\n", name, para_strerror(-ret));
		return ret;
	}
	id = *(uint32_t *)obj.data;
	return para_printf(&lbad->pb, "%u\t%s\n", id, name);
}

static void com_lsblob_callback(struct osl_table *table,
		int fd, const struct osl_object *query)
{
	struct lsblob_action_data lbad = {
		.flags = *(uint32_t *)query->data,
		.pb = {
			.max_size = SHMMAX,
			.private_data = &fd,
			.max_size_handler = pass_buffer_as_shm
		}
	};
	struct pattern_match_data pmd = {
		.table = table,
		.patterns = {.data = (char *)query->data + sizeof(uint32_t),
			.size = query->size - sizeof(uint32_t)},
		.pm_flags = PM_NO_PATTERN_MATCHES_EVERYTHING | PM_SKIP_EMPTY_NAME,
		.match_col_num = BLOBCOL_NAME,
		.data = &lbad,
		.action = print_blob,
	};
	int ret;

	if (lbad.flags & BLOB_LS_FLAG_REVERSE)
		pmd.pm_flags |= PM_REVERSE_LOOP;
	if (!(lbad.flags & BLOB_LS_FLAG_SORT_BY_ID))
		pmd.loop_col_num = BLOBCOL_NAME;
	else
		pmd.loop_col_num = BLOBCOL_ID;
	ret = for_each_matching_row(&pmd);
	if (ret < 0)
		para_printf(&lbad.pb, "%s\n", para_strerror(-ret));
	if (lbad.pb.offset)
		pass_buffer_as_shm(lbad.pb.buf, lbad.pb.offset, &fd);
	free(lbad.pb.buf);
}

static int com_lsblob(callback_function *f, int fd, int argc, char * const * const argv)
{
	uint32_t flags = 0;
	struct osl_object options = {.data = &flags, .size = sizeof(flags)};
	int i;

	for (i = 1; i < argc; i++) {
		const char *arg = argv[i];
		if (arg[0] != '-')
			break;
		if (!strcmp(arg, "--")) {
			i++;
			break;
		}
		if (!strcmp(arg, "-l")) {
			flags |= BLOB_LS_FLAG_LONG;
			continue;
		}
		if (!strcmp(arg, "-i")) {
			flags |= BLOB_LS_FLAG_SORT_BY_ID;
			continue;
		}
		if (!strcmp(arg, "-r")) {
			flags |= BLOB_LS_FLAG_REVERSE;
			continue;
		}
		break;
	}
//	if (argc > i)
//		return -E_BLOB_SYNTAX;
	return send_option_arg_callback_request(&options, argc - i,
		argv + i, f, send_result, &fd);
}

static int cat_blob(struct osl_table *table, struct osl_row *row,
		__a_unused const char *name, void *data)
{
	int ret = 0, ret2;
	struct osl_object obj;

	ret = osl_open_disk_object(table, row, BLOBCOL_DEF, &obj);
	if (ret < 0)
		return ret;
	if (obj.size)
		ret = pass_buffer_as_shm(obj.data, obj.size, data);
	ret2 = osl_close_disk_object(&obj);
	return (ret < 0)? ret : ret2;
}

static void com_catblob_callback(struct osl_table *table, int fd,
		const struct osl_object *query)
{
	struct pattern_match_data pmd = {
		.table = table,
		.patterns = *query,
		.loop_col_num = BLOBCOL_NAME,
		.match_col_num = BLOBCOL_NAME,
		.pm_flags = PM_SKIP_EMPTY_NAME,
		.data = &fd,
		.action = cat_blob
	};
	for_each_matching_row(&pmd);
}

static int com_catblob(callback_function *f, int fd, int argc,
		char * const * const argv)
{
	if (argc < 2)
		return -E_BLOB_SYNTAX;
	return send_standard_callback_request(argc - 1, argv + 1, f, send_result, &fd);
}

/** Used for removing rows from a blob table. */
struct rmblob_data {
	/** Message buffer. */
	struct para_buffer pb;
	/** Number of removed blobs. */
	unsigned num_removed;
};

static int remove_blob(struct osl_table *table, struct osl_row *row,
		const char *name, void *data)
{
	struct rmblob_data *rmbd = data;
	int ret = osl_del_row(table, row);
	if (ret < 0) {
		para_printf(&rmbd->pb, "%s: %s\n", name, para_strerror(-ret));
		return ret;
	}
	rmbd->num_removed++;
	return 1;
}

static void com_rmblob_callback(struct osl_table *table, int fd,
		const struct osl_object *query)
{
	int ret, ret2 = 0;
	struct rmblob_data rmbd = {
		.num_removed = 0,
		.pb = {
			.max_size = SHMMAX,
			.private_data = &fd,
			.max_size_handler = pass_buffer_as_shm
		}
	};
	struct pattern_match_data pmd = {
		.table = table,
		.patterns = *query,
		.loop_col_num = BLOBCOL_NAME,
		.match_col_num = BLOBCOL_NAME,
		.pm_flags = PM_SKIP_EMPTY_NAME,
		.data = &rmbd,
		.action = remove_blob
	};
	ret = for_each_matching_row(&pmd);
	if (ret < 0) {
		ret2 = para_printf(&rmbd.pb, "%s\n", para_strerror(-ret));
		if (ret2 < 0)
			goto out;
	}
	if (!rmbd.num_removed)
		ret2 = para_printf(&rmbd.pb, "no matches, nothing removed\n");
	else {
		ret2 = para_printf(&rmbd.pb, "removed %d blobs\n", rmbd.num_removed);
		afs_event(BLOB_RENAME, NULL, table);
	}
out:
	if (ret2 >= 0 && rmbd.pb.offset)
		pass_buffer_as_shm(rmbd.pb.buf, rmbd.pb.offset, &fd);
	free(rmbd.pb.buf);
}

static int com_rmblob(callback_function *f, int fd, int argc,
		char * const * const argv)
{
	if (argc < 2)
		return -E_MOOD_SYNTAX;
	return send_option_arg_callback_request(NULL, argc - 1, argv + 1, f,
		send_result, &fd);
}

static void com_addblob_callback(struct osl_table *table, __a_unused int fd,
		const struct osl_object *query)
{
	struct osl_object objs[NUM_BLOB_COLUMNS];
	char *name = query->data;
	size_t name_len = strlen(name) + 1;
	uint32_t id;
	unsigned num_rows;
	int ret;

	ret = osl_get_num_rows(table, &num_rows);
	if (ret < 0)
		goto out;
	if (!num_rows) { /* this is the first entry ever added */
		/* insert dummy row containing the id */
		id = 2; /* this entry will be entry #1, so 2 is the next */
		objs[BLOBCOL_ID].data = &id;
		objs[BLOBCOL_ID].size = sizeof(id);
		objs[BLOBCOL_NAME].data = "";
		objs[BLOBCOL_NAME].size = 1;
		objs[BLOBCOL_DEF].data = "";
		objs[BLOBCOL_DEF].size = 1;
		ret = osl_add_row(table, objs);
		if (ret < 0)
			goto out;
	} else {
		/* check if name already exists */
		struct osl_row *row;
		struct osl_object obj = {.data = name, .size = name_len};
		ret = osl_get_row(table, BLOBCOL_NAME, &obj, &row);
		if (ret < 0 && ret != -E_RB_KEY_NOT_FOUND)
			goto out;
		if (ret >= 0) { /* we already have a blob with this name */
			obj.data = name + name_len;
			obj.size = query->size - name_len;
			ret = osl_update_object(table, row, BLOBCOL_DEF, &obj);
			goto out;
		}
		/* new blob, get id of the dummy row and increment it */
		obj.data = "";
		obj.size = 1;
		ret = osl_get_row(table, BLOBCOL_NAME, &obj, &row);
		if (ret < 0)
			goto out;
		ret = osl_get_object(table, row, BLOBCOL_ID, &obj);
		if (ret < 0)
			goto out;
		id = *(uint32_t *)obj.data + 1;
		obj.data = &id;
		ret = osl_update_object(table, row, BLOBCOL_ID, &obj);
		if (ret < 0)
			goto out;
	}
	id--;
	objs[BLOBCOL_ID].data = &id;
	objs[BLOBCOL_ID].size = sizeof(id);
	objs[BLOBCOL_NAME].data = name;
	objs[BLOBCOL_NAME].size = name_len;
	objs[BLOBCOL_DEF].data = name + name_len;
	objs[BLOBCOL_DEF].size = query->size - name_len;
	ret = osl_add_row(table, objs);
	if (ret < 0)
		goto out;
	afs_event(BLOB_ADD, NULL, table);
out:
	if (ret < 0)
		PARA_NOTICE_LOG("%s\n", para_strerror(-ret));
}

static int com_addblob(callback_function *f, int fd, int argc,
		char * const * const argv)
{
	struct osl_object arg_obj;

	if (argc != 2)
		return -E_BLOB_SYNTAX;
	if (!*argv[1]) /* empty name is reserved for the dummy row */
		return -E_BLOB_SYNTAX;
	arg_obj.size = strlen(argv[1]) + 1;
	arg_obj.data = (char *)argv[1];
	return stdin_command(fd, &arg_obj, f, 10 * 1024 * 1024, NULL, NULL);
}

/* FIXME: Print output to client, not to log file */
static void com_mvblob_callback(struct osl_table *table, __a_unused int fd,
		const struct osl_object *query)
{
	char *src = (char *) query->data;
	struct osl_object obj = {.data = src, .size = strlen(src) + 1};
	char *dest = src + obj.size;
	struct osl_row *row;
	int ret = osl_get_row(table, BLOBCOL_NAME, &obj, &row);

	if (ret < 0)
		goto out;
	obj.data = dest;
	obj.size = strlen(dest) + 1;
	ret = osl_update_object(table, row, BLOBCOL_NAME, &obj);
	if (ret < 0)
		goto out;
	afs_event(BLOB_RENAME, NULL, table);
out:
	if (ret < 0)
		PARA_NOTICE_LOG("%s\n", para_strerror(-ret));
}

static int com_mvblob(callback_function *f, __a_unused int fd,
		int argc, char * const * const argv)
{
	if (argc != 3)
		return -E_MOOD_SYNTAX;
	return send_option_arg_callback_request(NULL, argc - 1, argv + 1, f,
		NULL, NULL);
}

#define DEFINE_BLOB_COMMAND(cmd_name, table_name, cmd_prefix) \
	static void com_ ## cmd_name ## cmd_prefix ## _callback(int fd, const struct osl_object *query) \
	{ \
		return com_ ## cmd_name ## blob_callback(table_name ## _table, fd, query); \
	} \
	int com_ ## cmd_name ## cmd_prefix(int fd, int argc, char * const * const argv) \
	{ \
		return com_ ## cmd_name ## blob(com_ ## cmd_name ## cmd_prefix ## _callback, fd, argc, argv); \
	}

static int blob_get_name_by_id(struct osl_table *table, uint32_t id,
		char **name)
{
	struct osl_row *row;
	struct osl_object obj = {.data = &id, .size = sizeof(id)};
	int ret;

	*name = NULL;
	if (!id)
		return 1;
	ret = osl_get_row(table, BLOBCOL_ID, &obj, &row);
	if (ret < 0)
		return ret;
	ret = osl_get_object(table, row, BLOBCOL_NAME, &obj);
	if (ret < 0)
		return ret;
	*name = (char *)obj.data;
	return 1;
}

/** Define the \p get_name_by_id function for this blob type. */
#define DEFINE_GET_NAME_BY_ID(table_name, cmd_prefix) \
	int cmd_prefix ## _get_name_by_id(uint32_t id, char **name) \
	{ \
		return blob_get_name_by_id(table_name ## _table, id, name); \
	}


static int blob_get_def_by_name(struct osl_table *table, char *name,
		struct osl_object *def)
{
	struct osl_row *row;
	struct osl_object obj = {.data = name, .size = strlen(name) + 1};
	int ret;

	def->data = NULL;
	if (!*name)
		return 1;
	ret = osl_get_row(table, BLOBCOL_NAME, &obj, &row);
	if (ret < 0)
		return ret;
	return osl_open_disk_object(table, row, BLOBCOL_DEF, def);
}

/** Define the \p get_def_by_id function for this blob type. */
#define DEFINE_GET_DEF_BY_NAME(table_name, cmd_prefix) \
	int cmd_prefix ## _get_def_by_name(char *name, struct osl_object *def) \
	{ \
		return blob_get_def_by_name(table_name ## _table, name, def); \
	}

static int blob_get_def_by_id(struct osl_table *table, uint32_t id,
		struct osl_object *def)
{
	struct osl_row *row;
	struct osl_object obj = {.data = &id, .size = sizeof(id)};
	int ret;

	def->data = NULL;
	if (!id)
		return 1;
	ret = osl_get_row(table, BLOBCOL_ID, &obj, &row);
	if (ret < 0)
		return ret;
	return osl_open_disk_object(table, row, BLOBCOL_DEF, def);
}

/** Define the \p get_def_by_id function for this blob type. */
#define DEFINE_GET_DEF_BY_ID(table_name, cmd_prefix) \
	int cmd_prefix ## _get_def_by_id(uint32_t id, struct osl_object *def) \
	{ \
		return blob_get_def_by_id(table_name ## _table, id, def); \
	}

static int blob_get_name_and_def_by_row(struct osl_table *table,
		const struct osl_row *row, char **name, struct osl_object *def)
{
	struct osl_object obj;
	int ret = osl_get_object(table, row, BLOBCOL_NAME, &obj);
	if (ret < 0)
		return ret;
	*name = obj.data;
	return osl_open_disk_object(table, row, BLOBCOL_DEF, def);
}
/** Define the \p get_name_and_def_by_row function for this blob type. */
#define DEFINE_GET_NAME_AND_DEF_BY_ROW(table_name, cmd_prefix) \
	int cmd_prefix ## _get_name_and_def_by_row(const struct osl_row *row, \
		char **name, struct osl_object *def) \
	{ \
		return blob_get_name_and_def_by_row(table_name ## _table, \
			row, name, def); \
	}

/** Define the \p close function for this blob type. */
#define DEFINE_BLOB_CLOSE(table_name) \
	void table_name ## _close(void) \
	{ \
		osl_close_table(table_name ## _table, OSL_MARK_CLEAN); \
		table_name ## _table = NULL; \
	}

/** Define the \p create function for this blob type. */
#define DEFINE_BLOB_CREATE(table_name) \
	int table_name ## _create(const char *dir) \
	{ \
		table_name ## _table_desc.dir = dir; \
		return osl_create_table(&table_name ## _table_desc); \
	}

static int blob_open(struct osl_table **table,
		struct osl_table_description *desc,
		const char *dir)
{
	int ret;
	desc->dir = dir;
	ret = osl_open_table(desc, table);
	if (ret >= 0)
		return ret;
	*table = NULL;
	if (ret >= 0 || is_errno(-ret, ENOENT))
		return 1;
	return ret;
}

#define DEFINE_BLOB_OPEN(table_name) \
	int table_name ## _open(const char *dir) \
	{ \
		return blob_open(&table_name ## _table, \
			&table_name ## _table_desc, dir); \
	}


/** Define the \p init function for this blob type. */
#define DEFINE_BLOB_INIT(table_name) \
	void table_name ## _init(struct afs_table *t) \
	{ \
		t->open = table_name ## _open; \
		t->close = table_name ## _close; \
		t->create = table_name ## _create;\
		t->event_handler = table_name ##_event_handler; \
		table_name ## _table = NULL; \
	}


/** Define all functions for this blob type. */
#define DEFINE_BLOB_FUNCTIONS(table_name, cmd_prefix) \
	DEFINE_BLOB_OPEN(table_name) \
	DEFINE_BLOB_CLOSE(table_name) \
	DEFINE_BLOB_CREATE(table_name) \
	DEFINE_BLOB_INIT(table_name) \
	DEFINE_BLOB_COMMAND(ls, table_name, cmd_prefix) \
	DEFINE_BLOB_COMMAND(cat, table_name, cmd_prefix) \
	DEFINE_BLOB_COMMAND(add, table_name, cmd_prefix) \
	DEFINE_BLOB_COMMAND(rm, table_name, cmd_prefix) \
	DEFINE_BLOB_COMMAND(mv, table_name, cmd_prefix) \
	DEFINE_GET_NAME_BY_ID(table_name, cmd_prefix); \
	DEFINE_GET_DEF_BY_ID(table_name, cmd_prefix); \
	DEFINE_GET_DEF_BY_NAME(table_name, cmd_prefix); \
	DEFINE_GET_NAME_AND_DEF_BY_ROW(table_name, cmd_prefix); \

/** \cond doxygen isn't smart enough to recognize these */
DEFINE_BLOB_FUNCTIONS(lyrics, lyr);
DEFINE_BLOB_FUNCTIONS(images, img);
DEFINE_BLOB_FUNCTIONS(moods, mood);
DEFINE_BLOB_FUNCTIONS(playlists, pl);
/** \endcond */
