/*
 * Copyright (C) 2005-2009 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file oggdec_filter.c Paraslash's ogg vorbis decoder. */

#include "para.h"

#include "oggdec_filter.cmdline.h"
#include "list.h"
#include "sched.h"
#include "ggo.h"
#include "filter.h"
#include "error.h"
#include "string.h"

#include <vorbis/vorbisfile.h>

/** Determine byte sex. */
#ifdef WORDS_BIGENDIAN
#define ENDIAN 1
#else
#define ENDIAN 0
#endif

/** Data specific to the oggdec filter. */
struct private_oggdec_data {
	/** Describes an ogg vorbis file. */
	OggVorbis_File *vf;
	/** The input buffer. */
	char *inbuf;
	/** The length of \a inbuf. */
	size_t inbuf_len;
	/** The number of bytes consumed from the input buffer. */
	size_t converted;
	/** When to start producing output. */
	struct timeval stream_start;
};

static size_t cb_read(void *buf, size_t size, size_t nmemb, void *datasource)
{
	struct filter_node *fn = datasource;
	struct private_oggdec_data *pod = fn->private_data;
	size_t ret, have = pod->inbuf_len - pod->converted;
	char *p = pod->inbuf + pod->converted;

//	PARA_DEBUG_LOG("pod = %p\n", pod);
//	PARA_DEBUG_LOG("vorbis requests %d bytes, have %d\n", size * nmemb, have);
	if (pod->inbuf_len < size) {
		if (*fn->fc->input_error)
			return 0;
		errno = EAGAIN;
		return (size_t)-1;
	}
	ret = PARA_MIN(nmemb, have / size) * size;
	memcpy(buf, p, ret);
	pod->converted += ret;
	return ret;
}

/*
 * Custom data seeking function.
 *
 * Since we want the data source to be treated as unseekable at all
 * times, the provided seek callback always returns -1 (failure).
 */
static int cb_seek(__a_unused void *datasource, __a_unused ogg_int64_t offset,
		__a_unused int whence)
{
	return -1;
}

static int cb_close(__a_unused void *datasource)
{
	return 0;
}

static const ov_callbacks ovc = {
	.read_func = cb_read,
	.seek_func = cb_seek,
	.close_func = cb_close,
	/*
	 * The tell function need not be provided if the data IO abstraction is
	 * not seekable
	 */
	.tell_func = NULL
};

static void ogg_open(struct filter_node *fn)
{
	struct private_oggdec_data *pod = para_calloc(
		sizeof(struct private_oggdec_data));
	struct oggdec_filter_args_info *conf = fn->conf;

	fn->private_data = pod;
	fn->bufsize = conf->bufsize_arg * 1024;
	fn->buf = para_malloc(fn->bufsize);
}

static void ogg_close(struct filter_node *fn)
{
	struct private_oggdec_data *pod = fn->private_data;
	if (pod->vf) {
		PARA_DEBUG_LOG("ov_clearing %p, pod = %p\n", pod->vf, pod);
		ov_clear(pod->vf);
		free(pod->vf);
		pod->vf = NULL;
	} else
		PARA_DEBUG_LOG("nothing to close in fc %p, pod = %p\n", pod->vf, pod);
	free(fn->buf);
	fn->buf = NULL;
	free(fn->private_data);
	fn->private_data = NULL;
}

static ssize_t ogg_convert(char *inbuffer, size_t len, struct filter_node *fn)
{
	ssize_t ret;
	struct private_oggdec_data *pod = fn->private_data;
	struct oggdec_filter_args_info *conf = fn->conf;
	/* make the buffer known to the read callback cb_read() */
	pod->inbuf = inbuffer;
	pod->inbuf_len = len;
	pod->converted = 0;

	if (!pod->vf) {
		if (*fn->fc->input_error < 0)
			return *fn->fc->input_error;
		if (!len)
			return 0;
		pod->vf = para_malloc(sizeof(struct OggVorbis_File));
		PARA_NOTICE_LOG("input buffer: %zd, opening ov callbacks\n", len);
		ret = ov_open_callbacks(fn, pod->vf,
			NULL, /* no initial buffer */
			0, /* no initial bytes */
			ovc); /* the ov_open_callbacks */
		if (ret == OV_ENOTVORBIS) {
			/* this might be due to the input buffer being too small */
			int ib = 1024 * conf->initial_buffer_arg; /* initial buffer */
			if (len < ib) {
				PARA_INFO_LOG("initial input buffer %zd/%d, "
					"waiting for more data\n", len, ib);
				free(pod->vf);
				pod->vf = NULL;
				return 0;
			}
			return -E_OGGDEC_NOTVORBIS;
		}
		if (ret == OV_EREAD)
			return -E_OGGDEC_READ;
		if (ret == OV_EVERSION)
			return -E_OGGDEC_VERSION;
		if (ret == OV_EBADHEADER)
			return -E_OGGDEC_BADHEADER;
		if (ret < 0)
			return -E_OGGDEC_FAULT;
		fn->fc->channels = ov_info(pod->vf, 0)->channels;
		fn->fc->samplerate = ov_info(pod->vf, 0)->rate;
		PARA_NOTICE_LOG("%d channels, %d Hz\n", fn->fc->channels,
			fn->fc->samplerate);
		/* wait a bit to avoid buffer underruns */
		tv_add(now, &(struct timeval){0, 500 * 1000}, &pod->stream_start);
		return pod->converted;
	}
	if (tv_diff(now, &pod->stream_start, NULL) < 0) {
		PARA_DEBUG_LOG("initial delay..\n");
		return 0;
	}
	while (fn->loaded < fn->bufsize) {
		int length = fn->bufsize - fn->loaded;
		long read_ret = ov_read(pod->vf, fn->buf + fn->loaded, length,
			ENDIAN, 2 /* 16 bit */, 1 /* signed */, NULL);
		if (read_ret == 0)
			return pod->converted;
		if (read_ret == OV_HOLE) {
			if (!fn->loaded) {
				PARA_INFO_LOG("hole, delaying playback\n");
				tv_add(now, &(struct timeval){0, 500 * 1000}, &pod->stream_start);
			}
			return pod->converted;
		}
		if (read_ret < 0)
			return -E_OGGDEC_BADLINK;
		fn->loaded += read_ret;
	}
	return pod->converted;
}

static int oggdec_parse_config(int argc, char **argv, void **config)
{
	int ret;
	struct oggdec_filter_args_info *ogg_conf;

	ogg_conf = para_calloc(sizeof(*ogg_conf));
	ret = -E_OGGDEC_SYNTAX;
	if (oggdec_cmdline_parser(argc, argv, ogg_conf))
		goto err;
	ret = -ERRNO_TO_PARA_ERROR(EINVAL);
	if (ogg_conf->bufsize_arg < 0)
		goto err;
	if (ogg_conf->bufsize_arg >= INT_MAX / 1024)
		goto err;
	if (ogg_conf->initial_buffer_arg < 0)
		goto err;
	if (ogg_conf->initial_buffer_arg >= INT_MAX / 1024)
		goto err;
	*config = ogg_conf;
	return 1;
err:
	free(ogg_conf);
	return ret;
}

/**
 * The init function of the ogg vorbis decoder.
 *
 * \param f Its fields are filled in by the function.
 */
void oggdec_filter_init(struct filter *f)
{
	struct oggdec_filter_args_info dummy;

	oggdec_cmdline_parser_init(&dummy);
	f->open = ogg_open;
	f->close = ogg_close;
	f->convert = ogg_convert;
	f->parse_config = oggdec_parse_config;
	f->help = (struct ggo_help) {
		.short_help = oggdec_filter_args_info_help,
		.detailed_help = oggdec_filter_args_info_detailed_help
	};
}
